import torch
import torch.nn as nn
import torchvision.transforms.functional as TF

class DoubleConv(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(DoubleConv, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, 3, 1, 1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, 3, 1, 1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        return self.conv(x)

class UNet(nn.Module):
    def __init__(
            self, in_channels=3, out_channels=1, features=[64, 128, 256, 512],
    ):
        super(UNet, self).__init__()
        self.ups = nn.ModuleList()
        self.downs = nn.ModuleList()
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)

        # Down part of UNET
        for feature in features:
            self.downs.append(DoubleConv(in_channels, feature))
            in_channels = feature

        # Up part of UNET
        for feature in reversed(features):
            self.ups.append(
                nn.ConvTranspose2d(
                    feature*2, feature, kernel_size=2, stride=2,
                )
            )
            self.ups.append(DoubleConv(feature*2, feature))

        self.bottleneck = DoubleConv(features[-1], features[-1]*2)
        self.final_conv = nn.Conv2d(features[0], out_channels, kernel_size=1)

    def forward(self, x):
        skip_connections = []

        for down in self.downs:
            x = down(x)
            skip_connections.append(x)
            x = self.pool(x)

        x = self.bottleneck(x)
        skip_connections = skip_connections[::-1]

        for idx in range(0, len(self.ups), 2):
            x = self.ups[idx](x)
            skip_connection = skip_connections[idx//2]

            if x.shape != skip_connection.shape:
                x = TF.resize(x, size=skip_connection.shape[2:])

            concat_skip = torch.cat((skip_connection, x), dim=1)
            x = self.ups[idx+1](concat_skip)

        return self.final_conv(x)


import torch
from torch.nn.utils.rnn import pad_sequence

from collections import Counter

class TextEncoder:
    def __init__(self, specials=["<unk>", "<pad>"], vocab=None,  preprocessor=None):
        """Initialize the TextProcessor with optional special tokens."""
        self.specials = specials
        self.vocab = vocab
        self.idx_to_token = None
        self.preprocessor = preprocessor

    def build_vocab(self, tokenized_texts):
        """Builds a vocabulary from tokenized texts."""
        all_tokens = [token for tokens in tokenized_texts for token in tokens]
        token_counts = Counter(all_tokens)
        self.vocab = {token: idx for idx, token in enumerate(self.specials + list(token_counts.keys()))}
        self.idx_to_token = {idx: token for token, idx in self.vocab.items()}

    def tokenize(self, text):
        """Tokenizes a text string."""
        if self.preprocessor is None:
            return text.lower().split()
        else:
            return self.preprocessor(text).split()

    def text_to_sequence(self, tokenized_text, unk_token="<unk>"):
        """Converts tokenized text to a sequence of indices based on the vocabulary."""
        return [self.vocab.get(token, self.vocab[unk_token]) for token in tokenized_text]

    def fit(self, texts, pad_token="<pad>"):
        """Converts a list of texts to padded sequences of indices."""
        tokenized_texts = [self.tokenize(text) for text in texts]

        # Build vocabulary if it hasn't been built yet
        if self.vocab is None:
            self.build_vocab(tokenized_texts)

        # Convert tokenized texts to sequences
        sequences = [torch.tensor(self.text_to_sequence(tokens), dtype=torch.long) for tokens in tokenized_texts]
        
        # Pad sequences
        pad_idx = self.vocab.get(pad_token, 0)  # Use 0 if pad_token not in vocab
        padded_sequences = pad_sequence(sequences, batch_first=True, padding_value=pad_idx)

        return padded_sequences

    def transform(self, texts, pad_token="<pad>"):
        """Converts new texts to sequences using the existing vocabulary."""
        if self.vocab is None:
            raise ValueError("Vocabulary not built. Please convert some texts first.")

        tokenized_texts = [self.tokenize(text) for text in texts]
        sequences = [torch.tensor(self.text_to_sequence(tokens), dtype=torch.long) for tokens in tokenized_texts]
        pad_idx = self.vocab.get(pad_token, 0)
        padded_sequences = pad_sequence(sequences, batch_first=True, padding_value=pad_idx)

        return padded_sequences

    def sequence_to_text(self, sequence, pad_token="<pad>"):
        """Converts a sequence of indices back to text."""
        return ' '.join([self.idx_to_token[idx] for idx in sequence if idx in self.idx_to_token and self.idx_to_token[idx] != pad_token])

    def inverse_transform(self, sequences, pad_token="<pad>"):
        """Converts a sequences of indices back to text using the index-to-token mapping."""
        return [self.sequence_to_text(sequence) for sequence in sequences]



from setuptools import setup, find_packages

setup(
    name='atanu_mlkit',
    version='0.0.3',
    author="Atanu Debnath",
    author_email="playatanu@gmail.com",
    description="Machine Learning Kit",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/playatanu/atanu-mlkit",
    project_urls={
        "Documentation": "https://playatanu.github.io/atanu-mlkit/",
        "Source": "https://github.com/playatanu/atanu-mlkit",
        "Tracker": "https://github.com/playatanu/atanu-mlkit/issues",
    },
    packages=find_packages(),
    license='MIT',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=[],
    python_requires='>=3.6',
)